"""
Tests for G90Alarm class
"""
import asyncio
from itertools import cycle
from unittest.mock import MagicMock
import pytest

from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.host_info import (
    G90HostInfo, G90HostInfoGsmStatus, G90HostInfoWifiStatus,
)
from pyg90alarm.local.host_status import (
    G90HostStatus,
)
from pyg90alarm.local.user_data_crc import (
    G90UserDataCRC,
)
from pyg90alarm.const import (
    G90RemoteButtonStates,
    G90RFIDKeypadStates,
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[100,[3,"PHONE","PRODUCT","206","206"]]IEND\0',
])
async def test_host_status(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving host status from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    res = await g90.get_host_status()

    assert await mock_device.recv_data == [b'ISTART[100,100,""]IEND\0']
    assert isinstance(res, G90HostStatus)
    assert res.host_status == 3
    assert res.product_name == 'PRODUCT'
    assert res.host_phone_number == 'PHONE'
    assert isinstance(res._asdict(), dict)


@pytest.mark.g90device(sent_data=[
    b'ISTART[206,'
    b'["DUMMYGUID","DUMMYPRODUCT",'
    b'"1.2","1.1","206","206",3,3,0,2,"4242",50,100]]IEND\0',
])
async def test_host_info(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving host information from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    res = await g90.get_host_info()

    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']
    assert isinstance(res, G90HostInfo)
    assert res.host_guid == 'DUMMYGUID'
    assert res.product_name == 'DUMMYPRODUCT'
    assert res.gsm_status == G90HostInfoGsmStatus.OPERATIONAL
    assert res.wifi_status == G90HostInfoWifiStatus.OPERATIONAL
    assert isinstance(res._asdict(), dict)


@pytest.mark.g90device(sent_data=[
    b'ISTART[160,["1","0xab","3","4","5","6"]]IEND\0',
    b'ISTART[160,["1","0xab","3","4","5","6"]]IEND\0',
])
async def test_user_data_crc(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving user data CRCs from the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    crc = await g90.get_user_data_crc()
    prop_crc = await g90.user_data_crc

    assert crc == prop_crc
    assert await mock_device.recv_data == [
        b'ISTART[160,160,""]IEND\0',
        b'ISTART[160,160,""]IEND\0',
    ]
    assert isinstance(crc, G90UserDataCRC)
    assert crc.sensor_list == '1'
    assert crc.device_list == '0xab'
    assert crc.history_list == '3'
    assert crc.scene_list == '4'
    assert crc.ifttt_list == '5'
    assert crc.fingerprint_list == '6'


# See `test_get_devices_concurrent` for the explanation of the test
@pytest.mark.g90device(sent_data=cycle([
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
]))
async def test_get_sensors_concurrent(mock_device: DeviceMock) -> None:
    """
    Tests for concurrently retrieving list of sensors produces consistent
    results.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    g90.paginated_result = MagicMock(  # type: ignore[method-assign]
        spec=g90.paginated_result, wraps=g90.paginated_result
    )

    res = await asyncio.gather(g90.get_sensors(), g90.get_sensors())
    assert g90.paginated_result.call_count == 2
    # pylint: disable=use-a-generator
    assert all([len(x) == 1 for x in res])


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,33,0,0,16,1,0,""]]]IEND\0',
    b'ISTART[102,'
    b'[[1,1,1],["Remote 2",11,0,10,1,0,33,0,0,16,1,0,""]]]IEND\0',
])
async def test_get_sensors_update(mock_device: DeviceMock) -> None:
    """
    Verifies updating the sensor list from the panel properly updates entitries
    exists in the list already, and marks those are not.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Ensure initial update results in single list entry
    sensors = await g90.get_sensors()
    assert len(sensors) == 1
    assert sensors[0].name == 'Remote'
    assert sensors[0].enabled is False
    # Check if existing enty is properly updated
    sensors = await g90.get_sensors()
    assert len(sensors) == 1
    assert sensors[0].name == 'Remote'
    assert sensors[0].enabled is True

    sensor_1 = sensors[0]

    # Ensure subsequent update results in two list entries, one being added and
    # another one is marked as unavailable (since it isn't present in the list
    # fetched from the device)
    sensors = await g90.get_sensors()
    assert len(sensors) == 2
    assert sensors[0].is_unavailable is True
    assert sensors[0].name == 'Remote'

    assert sensors[1].is_unavailable is False
    assert sensors[1].name == 'Remote 2'

    # Ensure the first entry is still in the list
    assert sensors[0] == sensor_1


@pytest.mark.g90device(sent_data=[
    b'ISTART[102,'
    b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTART[102,'
    b'[[1,1,1],["Remote 2",11,0,10,1,0,33,0,0,16,1,0,""]]]IEND\0',
])
async def test_find_sensor(mock_device: DeviceMock) -> None:
    """
    Verifies updating the sensor list from the panel properly updates entitries
    exists in the list already, and marks those are not.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    sensor = await g90.find_sensor(10, 'Remote')
    assert sensor is not None
    assert sensor.name == 'Remote'

    await g90.get_sensors()
    sensor = await g90.find_sensor(10, 'Remote')
    assert sensor is None
    sensor = await g90.find_sensor(10, 'Remote', exclude_unavailable=False)
    assert sensor is not None


@pytest.mark.parametrize('cache_alert_flags', [
    pytest.param(False, marks=pytest.mark.g90device(
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
            b'ISTART[117,[256]]IEND\0',
        ],
        notification_data=[
            b'[170,[5,[10,"Remote"]]]\0',
        ]
    ), id='Sensor activity callback'),
    pytest.param(True, marks=pytest.mark.g90device(
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
            b'ISTART[117,[256]]IEND\0',
            # Simulate wrong response retrieving alert configuration flags, so
            # that previously cached value is used
            b'ISTART[200,[256]]IEND\0',
        ],
        notification_data=[
            b'[170,[5,[10,"Remote"]]]\0',
        ]
    ), id='Sensor activity callback with cached alert flags'),
    pytest.param(False, marks=pytest.mark.g90device(
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Remote",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
            # Simulate wrong response retrieving alert configuration flags with
            # no previously cached data available
            b'ISTART[200,[256]]IEND\0',
        ],
        notification_data=[
            b'[170,[5,[10,"Remote"]]]\0',
        ]
    ), id='Sensor activity callback with alert flags resulting in error'),
])
async def test_sensor_callback(
    cache_alert_flags: bool, mock_device: DeviceMock
) -> None:
    """
    Tests for sensor callback.
    """
    reset_interval = 0.5
    g90 = G90Alarm(
        host=mock_device.host, port=mock_device.port,
        reset_occupancy_interval=reset_interval
    )
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )

    sensors = await g90.sensors

    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 10 and x.name == 'Remote']
    state_cb = MagicMock()
    state_cb.side_effect = lambda *args: future.set_result(True)
    sensor[0].state_callback = state_cb

    # Cache the alert flags if requested by the test
    if cache_alert_flags:
        await g90.alert_config.flags

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await g90.close_notifications()
    # Once passed this implies the G90Alarm sensor callback works as
    # expected, as it updates the occupancy states of the sensor
    assert sensor[0].occupancy

    # Re-create future since the callback will be invoked again when the
    # occupancy state is reset
    future = asyncio.get_running_loop().create_future()
    # Verify the occupancy state is reset upon configured interval
    await asyncio.sleep(reset_interval)
    assert not sensor[0].occupancy


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",26,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,26,1,4,"Remote","DUMMYGUID",1719223959,0,[""]]]\0',
        # Simulate sensor activity, which should reset low battery state for it
        b'[170,[5,[26,"Remote"]]]\0',
    ]
)
async def test_sensor_low_battery_callback(mock_device: DeviceMock) -> None:
    """
    Tests for sensor low battery callback.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )

    sensors = await g90.sensors

    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 26 and x.name == 'Remote']
    low_battery_sensor_cb = MagicMock()
    low_battery_sensor_cb.side_effect = lambda *args: future.set_result(True)
    sensor[0].low_battery_callback = low_battery_sensor_cb
    low_battery_cb = MagicMock()
    g90.low_battery_callback = low_battery_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    low_battery_sensor_cb.assert_called_once_with()
    low_battery_cb.assert_called_once_with(26, 'Remote')
    # Verify the low battery state is set upon receiving the notification
    assert sensor[0].is_low_battery is True

    # Signal the second notification is ready, the future has to be re-created
    # as the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    # Verify the low battery state is reset upon sensor activity
    assert sensor[0].is_low_battery is False

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Hall",21,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    ],
    notification_data=[
        b'[170,[6,[21,"Hall"]]]\0',
        b'[170,[1,[3]]]\0',
    ]
)
async def test_sensor_door_open_when_arming_callback(
    mock_device: DeviceMock
) -> None:
    """
    Tests for sensor door open when arming callback.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )

    sensors = await g90.sensors

    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 21 and x.name == 'Hall']
    door_open_when_arming_sensor_cb = MagicMock()
    door_open_when_arming_sensor_cb.side_effect = (
        lambda *args: future.set_result(True)
    )
    sensor[0].door_open_when_arming_callback = door_open_when_arming_sensor_cb
    door_open_when_arming_cb = MagicMock()
    g90.door_open_when_arming_callback = door_open_when_arming_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    door_open_when_arming_sensor_cb.assert_called_once_with()
    door_open_when_arming_cb.assert_called_once_with(21, 'Hall')
    # Verify the door open when arming state is set upon receiving the
    # notification
    assert sensor[0].is_door_open_when_arming is True

    # Signal the second notification is ready, the future has to be re-created
    # as the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    # Verify the door open when arming state is reset upon disarming
    assert sensor[0].is_door_open_when_arming is False

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[3,1,3],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Cord 1",12,0,126,1,0,32,0,5,16,1,0,""]'
        b']]IEND\0',
    ],
    notification_data=[
        b'[170,[1,[1]]]\0'
    ]
)
async def test_armdisarm_callback(mock_device: DeviceMock) -> None:
    """
    Tests for arm/disarm callback.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.armdisarm_callback = armdisarm_cb
    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await g90.close_notifications()
    armdisarm_cb.assert_called_once_with(1)


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Door",100,0,1,1,0,32,0,0,16,1,0,""]]]IEND\0',
        # The GETNOTICEFLAG command is invoked twice, for each of
        # alerts simulated below
        b'ISTART[117,[256]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,100,1,1,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
        b'[208,[4,100,1,0,"Door","DUMMYGUID",1631545189,0,[""]]]\0',
    ]
)
async def test_door_open_close_callback(mock_device: DeviceMock) -> None:
    """
    Tests for door open/close callback.
    """
    future = asyncio.get_running_loop().create_future()
    door_open_close_cb = MagicMock()
    door_open_close_cb.side_effect = lambda *args: future.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.door_open_close_callback = door_open_close_cb

    # Simulate two device alerts - for opening (this one) and then closing the
    # door (see below)
    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Corresponding sensor should turn to occupied (=door opened)
    sensors = await g90.sensors
    assert sensors[0].occupancy

    # Simulate second device alert for door close
    await mock_device.send_next_notification()
    # Signal the second alert is ready, the future has to be re-created as
    # the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()

    await asyncio.wait([future], timeout=0.1)
    # The sensor should become inactive (=door closed)
    sensors = await g90.sensors
    assert not sensors[0].occupancy

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        # Simulate panel with two sensors
        b'ISTART[102,'
        b'[[3,1,3],'
        b'["Hall",100,0,1,1,0,32,0,0,16,1,0,""],'
        b'["Room",101,0,1,1,0,32,0,0,16,1,0,""],'
        b'["Infrared",103,0,8,1,0,32,0,0,16,1,0,""]'
        b']]IEND\0',
        # Alert configuration, used by sensor activity callback invoked when
        # handling alarm
        b'ISTART[117,[256]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[3,100,1,1,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
        b'[208,[3,101,1,1,"Room","DUMMYGUID",1630876128,0,[""]]]\0',
        b'[208,[3,102,1,1,"No Room","DUMMYGUID",1630876128,0,[""]]]\0',
        b'[208,[3,103,8,0,"Infrared","DUMMYGUID",1630876128,0,[""]]]\0',
    ]
)
async def test_alarm_callback(mock_device: DeviceMock) -> None:
    """
    Tests for alarm callback.
    """
    future = asyncio.get_running_loop().create_future()
    alarm_cb = MagicMock()
    alarm_cb.side_effect = lambda *args: future.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    sensors = await g90.sensors
    # Set extra data for the 1st sensor
    sensors[0].extra_data = 'Dummy extra data'

    g90.alarm_callback = alarm_cb
    await g90.listen_notifications()
    # Simulate three alarm notifications - for sensor with extra data,
    # another for sensor with no extra data, and third for non-existent
    # sensor
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Verify extra data is passed to the callback
    alarm_cb.assert_called_once_with(100, 'Hall', 'Dummy extra data')
    # Verify the triggering sensor is set to active
    assert sensors[0].occupancy is True

    # Simulate alarm for sensor with no extra data
    alarm_cb.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Verify no extra data is passed to the callback
    alarm_cb.assert_called_once_with(101, 'Room', None)
    # Verify the triggering sensor is set to active
    assert sensors[0].occupancy is True

    # Simulate alarm for non-existent sensor
    alarm_cb.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Simulate callback is called with no data
    alarm_cb.assert_called_once_with(102, 'No Room', None)

    # Simulate alarm for infrared sensor
    alarm_cb.reset_mock()
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    # Simulate callback is called with no data
    alarm_cb.assert_called_once_with(103, 'Infrared', None)
    # Verify the triggering sensor is set to active
    assert sensors[2].occupancy is True

    await g90.close_notifications()


@pytest.mark.parametrize([], [
    pytest.param(marks=pytest.mark.g90device(
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Hall",100,0,1,1,0,32,0,0,16,1,0,""]]]IEND\0',
            # Alert configuration, used by sensor activity callback invoked
            # when handling alarm
            b'ISTART[117,[256]]IEND\0',
        ],
        notification_data=[
            b'[208,[3,100,1,3,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
            b'[170,[1,[3]]]\0',
        ]
    ), id='Regular sensor tamper alert'),
    pytest.param(marks=pytest.mark.g90device(
        sent_data=[
            b'ISTART[102,'
            b'[[1,1,1],["Hall",100,0,8,1,0,32,0,0,16,1,0,""]]]IEND\0',
            # Same as above
            b'ISTART[117,[256]]IEND\0',
        ],
        notification_data=[
            b'[208,[3,100,8,1,"Hall","DUMMYGUID",1630876128,0,[""]]]\0',
            b'[170,[1,[3]]]\0',
        ]
    ), id='Infrared sensor tamper alert'),
])
async def test_sensor_tamper_callback(
    mock_device: DeviceMock
) -> None:
    """
    Tests for sensor tamper callback.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )

    sensors = await g90.sensors

    future = asyncio.get_running_loop().create_future()
    sensor = [x for x in sensors if x.index == 100 and x.name == 'Hall']
    tamper_sensor_cb = MagicMock()
    tamper_sensor_cb.side_effect = (
        lambda *args: future.set_result(True)
    )
    sensor[0].tamper_callback = tamper_sensor_cb
    tamper_cb = MagicMock()
    g90.tamper_callback = tamper_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    tamper_sensor_cb.assert_called_once_with()
    tamper_cb.assert_called_once_with(100, 'Hall')
    # Verify the sensor tampered state is set upon receiving the
    # notification
    assert sensor[0].is_tampered is True

    # Signal the second notification is ready, the future has to be re-created
    # as the corresponding callback will be fired again
    future = asyncio.get_running_loop().create_future()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)

    # Verify the sensor tampered state is reset upon disarming
    assert sensor[0].is_tampered is False

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",11,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        # Host SOS
        b'[208,[1,1,0,0,"","DUMMYGUID",1734175050,0,[""]]]\0',
        # Remote SOS
        b'[208,[3,11,10,3,"Remote","DUMMYGUID",1734177048,0,[""]]]\0',
    ]
)
async def test_sos_callback(mock_device: DeviceMock) -> None:
    """
    Tests for SOS callback.
    """
    future_sos = asyncio.get_running_loop().create_future()
    future_alarm = asyncio.get_running_loop().create_future()

    sos_cb = MagicMock()
    sos_cb.side_effect = lambda *args: future_sos.set_result(True)
    alarm_cb = MagicMock()
    alarm_cb.side_effect = lambda *args: future_alarm.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.sos_callback = sos_cb
    g90.alarm_callback = alarm_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sos, future_alarm], timeout=0.1)
    sos_cb.assert_called_once_with(1, 'Host SOS', True)
    alarm_cb.assert_called_once_with(1, 'Host SOS', None)

    sos_cb.reset_mock()
    alarm_cb.reset_mock()
    future_sos = asyncio.get_running_loop().create_future()
    future_alarm = asyncio.get_running_loop().create_future()
    future_button = asyncio.get_running_loop().create_future()
    button_cb = MagicMock()
    button_cb.side_effect = lambda *args: future_button.set_result(True)
    g90.remote_button_press_callback = button_cb
    await mock_device.send_next_notification()
    await asyncio.wait([future_sos, future_alarm, future_button], timeout=0.1)
    sos_cb.assert_called_once_with(11, 'Remote', False)
    alarm_cb.assert_called_once_with(11, 'Remote', None)
    # Button press callback should be called with the remote button state, but
    # only for SOS initiated by the remote
    button_cb.assert_called_once_with(11, 'Remote', G90RemoteButtonStates.SOS)
    await g90.close_notifications()

    # Test incurs other callbacks and corresponding tasks (two invocations of
    # on_sensor_activity callback for reflecting state changes for the remote
    # as a sensor), but those aren't in scope - explicitly await any
    # pending tasks to avoid asyncio warnings
    await asyncio.gather(
        *(t for t in asyncio.all_tasks() if t is not asyncio.current_task()),
        # Ignore exceptions from other tasks
        return_exceptions=True
    )


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Remote",11,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,11,10,0,"Remote","DUMMYGUID",1734176900,0,[""]]]\0',
    ]
)
async def test_remote_button_callback(mock_device: DeviceMock) -> None:
    """
    Tests for remote button callback.
    """
    future_sensor = asyncio.get_running_loop().create_future()
    future_button = asyncio.get_running_loop().create_future()
    sensor_cb = MagicMock()
    sensor_cb.side_effect = lambda *args: future_sensor.set_result(True)
    button_cb = MagicMock()
    button_cb.side_effect = lambda *args: future_button.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.sensor_callback = sensor_cb
    g90.remote_button_press_callback = button_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sensor, future_button], timeout=0.1)
    sensor_cb.assert_called_once_with(11, 'Remote', True)
    button_cb.assert_called_once_with(
        11, 'Remote', G90RemoteButtonStates.ARM_AWAY
    )

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Keypad-",15,0,11,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,15,11,7,"Keypad-","DUMMYGUID",1734176900,0,[""]]]\0',
    ]
)
async def test_rfid_callback(mock_device: DeviceMock) -> None:
    """
    Tests for RFID keypad callback.
    """
    future_sensor = asyncio.get_running_loop().create_future()
    future_keypad = asyncio.get_running_loop().create_future()
    sensor_cb = MagicMock()
    sensor_cb.side_effect = lambda *args: future_sensor.set_result(True)
    keypad_cb = MagicMock()
    keypad_cb.side_effect = lambda *args: future_keypad.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.sensor_callback = sensor_cb
    g90.rfid_keypad_callback = keypad_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sensor, future_keypad], timeout=0.1)
    sensor_cb.assert_called_once_with(15, 'Keypad-', True)
    keypad_cb.assert_called_once_with(
        15, 'Keypad-', G90RFIDKeypadStates.CARD_1
    )

    await g90.close_notifications()


@pytest.mark.g90device(
    sent_data=[
        b'ISTART[102,'
        b'[[1,1,1],["Keypad-",15,0,11,1,0,32,0,0,16,1,0,""]]]IEND\0',
        b'ISTART[117,[256]]IEND\0',
    ],
    notification_data=[
        b'[208,[4,15,11,5,"Keypad-","DUMMYGUID",1734176900,0,[""]]]\0',
    ]
)
async def test_rfid_low_battery_callback(mock_device: DeviceMock) -> None:
    """
    Tests for RFID keypad low battery callback.
    """
    future_sensor = asyncio.get_running_loop().create_future()
    future_keypad = asyncio.get_running_loop().create_future()
    sensor_cb = MagicMock()
    low_battery_cb = MagicMock()
    low_battery_cb.side_effect = lambda *args: future_sensor.set_result(True)
    keypad_cb = MagicMock()
    keypad_cb.side_effect = lambda *args: future_keypad.set_result(True)

    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.sensor_callback = sensor_cb
    g90.rfid_keypad_callback = keypad_cb
    g90.low_battery_callback = low_battery_cb

    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future_sensor, future_keypad], timeout=0.1)
    # Sensor callback should not be called for low battery notification on RFID
    # keypad
    sensor_cb.assert_not_called()
    # Both low battery and keypad callbacks should be called
    low_battery_cb.assert_called_once_with(15, 'Keypad-')
    keypad_cb.assert_called_once_with(
        15, 'Keypad-', G90RFIDKeypadStates.LOW_BATTERY
    )

    await g90.close_notifications()


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_arm_away(mock_device: DeviceMock) -> None:
    """
    Tests for arming the device in away mode.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.arm_away()
    assert await mock_device.recv_data == [
        b'ISTART[101,101,[101,[1]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_arm_home(mock_device: DeviceMock) -> None:
    """
    Tests for arming the device in home mode.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.arm_home()
    assert await mock_device.recv_data == [
        b'ISTART[101,101,[101,[2]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_disarm(mock_device: DeviceMock) -> None:
    """
    Tests for disarming the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.disarm()
    assert await mock_device.recv_data == [
        b'ISTART[101,101,[101,[3]]]IEND\0',
    ]


@pytest.mark.g90device(
    sent_data=[
        # Get alert configuration
        b"ISTART[117,[1]]IEND\0",
        b"ISTARTIEND\0",
        # Simulated list of sensors, which is used to reset door open when
        # arming/tamper flags on those had the flags set when arming
        b'ISTART[102,'
        b'[[3,1,3],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Cord 1",12,0,126,1,0,32,0,5,16,1,0,""]'
        b']]IEND\0',
    ],
    notification_data=[
        b'[170,[1,[1]]]\0',
    ]
)
async def test_sms_alert_when_armed(mock_device: DeviceMock) -> None:
    """
    Tests for enabling SMS alerts when device is armed.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.armdisarm_callback = armdisarm_cb
    g90.sms_alert_when_armed = True
    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await g90.close_notifications()
    assert set([
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[513]]]IEND\0",
    ]).issubset(set(await mock_device.recv_data))


@pytest.mark.g90device(
    sent_data=[
        # See above for the clarification on command sequence
        b"ISTART[117,[513]]IEND\0",
        b"ISTARTIEND\0",
        b'ISTART[102,'
        b'[[3,1,3],["Remote 1",10,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Remote 2",11,0,10,1,0,32,0,0,16,1,0,""],'
        b'["Cord 1",12,0,126,1,0,32,0,5,16,1,0,""]'
        b']]IEND\0',
    ],
    notification_data=[
        b'[170,[1,[3]]]\0',
    ]
)
async def test_sms_alert_when_disarmed(mock_device: DeviceMock) -> None:
    """
    Tests for disabling SMS alerts when device is disarmed.
    """
    future = asyncio.get_running_loop().create_future()
    armdisarm_cb = MagicMock()
    armdisarm_cb.side_effect = lambda *args: future.set_result(True)
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    await g90.use_local_notifications(
        notifications_local_host=mock_device.notification_host,
        notifications_local_port=mock_device.notification_port
    )
    g90.armdisarm_callback = armdisarm_cb
    g90.sms_alert_when_armed = True
    await g90.listen_notifications()
    await mock_device.send_next_notification()
    await asyncio.wait([future], timeout=0.1)
    await g90.close_notifications()
    assert set([
        b'ISTART[117,117,""]IEND\0',
        b'ISTART[117,117,""]IEND\0',
        b"ISTART[116,116,[116,[1]]]IEND\0",
    ]).issubset(set(await mock_device.recv_data))



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Protocol entity for G90 alarm panel config.
"""
from __future__ import annotations
from typing import Dict, Any
from dataclasses import dataclass
from enum import IntEnum
from .dataclass_load_save import DataclassLoadSave
from ..const import G90Commands


class G90SpeechLanguage(IntEnum):
    """
    Supported speech languages.
    """
    ENGLISH_FEMALE = 1
    ENGLISH_MALE = 2
    CHINESE_FEMALE = 3
    CHINESE_MALE = 4
    GERMAN_FEMALE = 5
    GERMAN_MALE = 6
    SPANISH_FEMALE = 7
    SPANISH_MALE = 8
    DUTCH_FEMALE = 9
    DUTCH_MALE = 10
    SWEDISH_FEMALE = 11
    SWEDISH_MALE = 12
    FRENCH_FEMALE = 13
    FRENCH_MALE = 14
    TURKISH_FEMALE = 15
    TURKISH_MALE = 16
    RUSSIAN_FEMALE = 17
    RUSSIAN_MALE = 18


class G90VolumeLevel(IntEnum):
    """
    Supported volume levels.
    """
    MUTE = 0
    LOW = 1
    HIGH = 2


@dataclass
class G90HostConfig(DataclassLoadSave):
    """
    Interprets data fields of GETHOSTCONFIG/SETHOSTCONFIG commands.
    """
    # pylint: disable=too-many-instance-attributes
    LOAD_COMMAND = G90Commands.GETHOSTCONFIG
    SAVE_COMMAND = G90Commands.SETHOSTCONFIG

    # Duration of the alarm siren when triggered, in seconds
    alarm_siren_duration: int
    # Delay before arming the panel, in seconds
    arm_delay: int
    # Delay before the alarm is triggered, in seconds
    alarm_delay: int
    # Duration of the backlight, in seconds
    backlight_duration: int
    # Alarm volume level, applies to panel's built-in speaker
    _alarm_volume_level: int
    # Speech volume level
    _speech_volume_level: int
    # Duration of the ring for the incoming call, in seconds
    ring_duration: int
    # Speech language
    _speech_language: int
    # Key tone volume level
    _key_tone_volume_level: int
    # Timezone offset, in minutes
    timezone_offset_m: int
    # Ring volume level for incoming calls
    _ring_volume_level: int

    @property
    def speech_language(self) -> G90SpeechLanguage:
        """
        Returns the speech language as an enum.
        """
        return G90SpeechLanguage(self._speech_language)

    @speech_language.setter
    def speech_language(self, value: G90SpeechLanguage) -> None:
        self._speech_language = value.value

    @property
    def alarm_volume_level(self) -> G90VolumeLevel:
        """
        Returns the alarm volume level as an enum.
        """
        return G90VolumeLevel(self._alarm_volume_level)

    @alarm_volume_level.setter
    def alarm_volume_level(self, value: G90VolumeLevel) -> None:
        self._alarm_volume_level = value.value

    @property
    def speech_volume_level(self) -> G90VolumeLevel:
        """
        Returns the speech volume level as an enum.
        """
        return G90VolumeLevel(self._speech_volume_level)

    @speech_volume_level.setter
    def speech_volume_level(self, value: G90VolumeLevel) -> None:
        self._speech_volume_level = value.value

    @property
    def key_tone_volume_level(self) -> G90VolumeLevel:
        """
        Returns the key tone volume level as an enum.
        """
        return G90VolumeLevel(self._key_tone_volume_level)

    @key_tone_volume_level.setter
    def key_tone_volume_level(self, value: G90VolumeLevel) -> None:
        self._key_tone_volume_level = value.value

    @property
    def ring_volume_level(self) -> G90VolumeLevel:
        """
        Returns the ring volume level as an enum.
        """
        return G90VolumeLevel(self._ring_volume_level)

    @ring_volume_level.setter
    def ring_volume_level(self, value: G90VolumeLevel) -> None:
        self._ring_volume_level = value.value

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the dataclass fields as a dictionary.
        """
        return {
            'alarm_siren_duration': self.alarm_siren_duration,
            'arm_delay': self.arm_delay,
            'alarm_delay': self.alarm_delay,
            'backlight_duration': self.backlight_duration,
            'alarm_volume_level': self.alarm_volume_level.name,
            'speech_volume_level': self.speech_volume_level.name,
            'ring_duration': self.ring_duration,
            'speech_language': self.speech_language.name,
            'key_tone_volume_level': self.key_tone_volume_level.name,
            'timezone_offset_m': self.timezone_offset_m,
            'ring_volume_level': self.ring_volume_level.name,
        }



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Defines mapping of various events.
"""
from __future__ import annotations
from typing import Dict, TYPE_CHECKING
import logging
from .const import (
    G90AlertSources,
    G90AlertStates,
    G90CommonSensorAlertStates,
    G90InfraredAlertStates,
)

_LOGGER = logging.getLogger(__name__)


def map_alert_state(
    source: G90AlertSources, state: int
) -> G90AlertStates:
    """
    Converts alert state of infrared and other sensors to consolidated one.

    :param source: Source of the alert
    :param state: State code as received from the panel
    :return: Consolidated alert state
    """
    if TYPE_CHECKING:
        # Forward declarations
        key: G90CommonSensorAlertStates | G90InfraredAlertStates
        mapping: Dict[
            G90CommonSensorAlertStates | G90InfraredAlertStates,
            G90AlertStates
        ]

    # Mapping for common sensors
    mapping = {
        G90CommonSensorAlertStates.DOOR_CLOSE:
            G90AlertStates.DOOR_CLOSE,
        G90CommonSensorAlertStates.DOOR_OPEN:
            G90AlertStates.DOOR_OPEN,
        G90CommonSensorAlertStates.SOS:
            G90AlertStates.SOS,
        G90CommonSensorAlertStates.TAMPER:
            G90AlertStates.TAMPER,
        G90CommonSensorAlertStates.LOW_BATTERY:
            G90AlertStates.LOW_BATTERY,
        G90CommonSensorAlertStates.ALARM:
            G90AlertStates.ALARM,
    }
    key = G90CommonSensorAlertStates(state)
    mapping_kind = 'common'

    # Mapping for infrared sensors
    if source == G90AlertSources.INFRARED:
        mapping = {
            G90InfraredAlertStates.MOTION_DETECTED:
                G90AlertStates.MOTION_DETECTED,
            G90InfraredAlertStates.TAMPER:
                G90AlertStates.TAMPER,
            G90InfraredAlertStates.LOW_BATTERY:
                G90AlertStates.LOW_BATTERY,
        }
        key = G90InfraredAlertStates(state)
        mapping_kind = 'infrared'

    try:
        result = mapping[key]
    except KeyError as exc:
        # Raise the error similar to Enum if state is invalid
        raise ValueError(
            f'{state} is not valid for source {source}'
        ) from exc

    _LOGGER.debug(
        'Mapped %s sensor alert state %d to consolidated state %s'
        ' for source %s',
        mapping_kind, state, repr(result), repr(source)
    )
    return result



# Copyright (c) 2025 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Device (relays, sockets) definitions for G90 alarm panel.
"""
from __future__ import annotations
import logging
from .base import (
    G90PeripheralDefinition,
    G90PeripheralDefinitionsBase,
    G90PeripheralTypes,
    G90PeripheralProtocols,
    G90PeripheralRwModes,
    G90PeripheralMatchModes,
    unique_definitions,
)

_LOGGER = logging.getLogger(__name__)


@unique_definitions
class G90DeviceDefinitions(G90PeripheralDefinitionsBase):
    """
    Device definitions, required when modifying them since
    writing a device to the panel requires values not present on read.
    """
    DEVICE_DEFINITIONS = [
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_DEV,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=3,
            rx=0,
            tx=2,
            private_data='060A0600',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: S07",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1190,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=0,
            rx=0,
            tx=2,
            private_data='0707070B0B0D0D0E0E00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: JDQ",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=4
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=1,
            rx=0,
            tx=2,
            private_data='07070700',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: Single channel",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=4,
            rx=0,
            tx=2,
            private_data='050D0500',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket Switch",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=840,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=5,
            rx=0,
            tx=2,
            private_data='070A0E080D060B00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: 3 channel",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=3
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET,
            subtype=6,
            rx=0,
            tx=2,
            private_data='0B0D0E0B0C090A070800',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: 4 channel",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=4
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CURTAIN,
            subtype=0,
            rx=0,
            tx=7,
            private_data='070B0E0D0C09030100',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Curtain: Rolling",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CURTAIN,
            subtype=1,
            rx=0,
            tx=7,
            private_data='0804010200',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Curtain",
            protocol=G90PeripheralProtocols.RF_SLIDER,
            timeout=0,
            baudrate=500,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CURTAIN,
            subtype=1,
            rx=0,
            tx=7,
            private_data='0E0B0E0D0C09030100',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Curtain: Sliding",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CURTAIN,
            subtype=2,
            rx=0,
            tx=7,
            private_data='070B0E0D0C09030100',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Curtain: Push",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.AIRCON,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Air Conditioner",
            protocol=G90PeripheralProtocols.RF_PRIVATE,
            timeout=0,
            baudrate=800,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.TV,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="TV",
            protocol=G90PeripheralProtocols.RF_PRIVATE,
            timeout=0,
            baudrate=800,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN,
            subtype=0,
            rx=0,
            tx=2,
            private_data='FEFEF600',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: SS08",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN,
            subtype=4,
            rx=0,
            tx=2,
            private_data='FEFEF600',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: SS07A",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN,
            subtype=5,
            rx=0,
            tx=2,
            private_data='FCFCCF00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: SS04",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN,
            subtype=1,
            rx=0,
            tx=2,
            private_data='FCFCCF00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: SS02B",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN,
            subtype=2,
            rx=0,
            tx=2,
            private_data='FEFEFD00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: Solar",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=2200,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.NIGHTLIGHT,
            subtype=0,
            rx=0,
            tx=2,
            private_data='060A0600',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Night Light",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1190,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOCKET_2_4G,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Socket: 2.4G",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SIREN_2_4G,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Siren: 2.4G",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=2200,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CURTAIN_2_4G,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Curtain: 2.4G",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
    ]

    @classmethod
    def definitions(cls) -> list[G90PeripheralDefinition]:
        """
        Gets all device definitions.

        :return: List of device definitions.
        """
        _LOGGER.debug(
            "Number of sensor definitions: %d", len(cls.DEVICE_DEFINITIONS)
        )
        return cls.DEVICE_DEFINITIONS



"""
Tests for the G90Alarm class and its methods related to devices.
"""
import asyncio
from itertools import cycle
from contextlib import nullcontext, AbstractContextManager
from unittest.mock import MagicMock
import pytest
from pyg90alarm import (
    G90Alarm, G90Device, G90EntityRegistrationError,
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
])
async def test_devices(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving devices from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]
    assert len(devices) == 1
    assert isinstance(devices, list)
    assert isinstance(devices[0], G90Device)
    assert devices[0].name == 'Switch'
    assert devices[0].index == 10
    assert isinstance(devices[0]._asdict(), dict)


# Provide an endless sequence of simulated panel responses for the devices
# list. Each attempt will simulate a single device. This sequence will prevent
# `G90TimeoutError` if the code under test initiates more exchanges with the
# panel than the simulated data contains.
@pytest.mark.g90device(sent_data=cycle([
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
]))
async def test_get_devices_concurrent(mock_device: DeviceMock) -> None:
    """
    Tests for concurrently retrieving list of devices produces consistent
    results.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    g90.paginated_result = MagicMock(  # type: ignore[method-assign]
        spec=g90.paginated_result, wraps=g90.paginated_result
    )

    # Issue two concurrent requests to retrieve devices
    res = await asyncio.gather(g90.get_devices(), g90.get_devices())
    # Ensure only corresponding number of exchanges with the panel
    assert g90.paginated_result.call_count == 2
    # While `pylint` demands use of generator, the comprehension is used
    # instead for ease of troubleshooting test failures as it will show the
    # list elements, not just generator instance
    # pylint: disable=use-a-generator
    assert all([len(x) == 1 for x in res])


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,2,0,""]]]IEND\0'
])
async def test_multinode_device(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving multi-node devices (e.g. multi-channel switch) from
    the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]
    assert len(devices) == 2
    assert isinstance(devices, list)
    assert isinstance(devices[0], G90Device)
    assert devices[0].name == 'Switch#1'
    assert devices[0].index == 10
    assert devices[0].subindex == 0
    assert devices[1].name == 'Switch#2'
    assert devices[1].index == 10
    assert devices[1].subindex == 1


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTARTIEND\0',
    b'ISTARTIEND\0',
])
async def test_control_device(mock_device: DeviceMock) -> None:
    """
    Tests for controlling devices from the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices

    assert devices == prop_devices
    await devices[0].turn_on()
    await devices[0].turn_off()
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
        b'ISTART[137,137,[137,[10,0,0]]]IEND\0',
        b'ISTART[137,137,[137,[10,1,0]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Water",10,0,7,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_device_unsupported_disable(mock_device: DeviceMock) -> None:
    """
    Tests for disabling an unsupported device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    prop_devices = await g90.devices
    assert devices == prop_devices
    assert devices[0].enabled
    await devices[0].set_enabled(False)
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Water",10,0,7,0,0,33,0,0,17,1,0,""]'
    b']]IEND\0',
    b"ISTARTIEND\0",
])
async def test_device_delete(mock_device: DeviceMock) -> None:
    """
    Tests for deleting the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    devices = await g90.get_devices()
    await devices[0].delete()
    assert devices[0].is_unavailable
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
        b'ISTART[136,136,[136,[10]]]IEND\0',
    ]


@pytest.mark.parametrize(
    'expected_exception', [
        # Normal operation, no exception expected
        pytest.param(
            nullcontext(),
            id='register-device-ok',
            marks=pytest.mark.g90device(sent_data=[
                b'ISTART[138,'
                b'[[1,1,1],'
                b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""]'
                b']]IEND\0',
                b'ISTARTIEND\0',
                b'ISTART[162,[0]]IEND\0',
                b'ISTART[138,'
                b'[[1,1,2],'
                b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""],'
                b'["Test socket",0,0,128,3,0,32,0,0,16,1,0,""]'
                b']]IEND\0',
            ]),
        ),
        # Operation simulating the panel responded with no index for the new
        # device
        pytest.param(
            pytest.raises(G90EntityRegistrationError),
            id='register-device-no-index-received',
            marks=pytest.mark.g90device(sent_data=[
                b'ISTART[138,'
                b'[[1,1,1],'
                b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""]'
                b']]IEND\0',
                b'ISTARTIEND\0',
                b'ISTARTIEND\0',
            ]),
        ),
        # Operation simulating panel responded with an index for the new
        # device, but such entity isn't found in the devices list
        pytest.param(
            pytest.raises(G90EntityRegistrationError),
            id='register-device-not-found',
            marks=pytest.mark.g90device(sent_data=[
                b'ISTART[138,'
                b'[[1,1,1],'
                b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""]'
                b']]IEND\0',
                b'ISTARTIEND\0',
                b'ISTART[162,[999]]IEND\0',
                b'ISTART[138,'
                b'[[1,1,2],'
                b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""],'
                b'["Test socket",0,0,128,3,0,32,0,0,16,1,0,""]'
                b']]IEND\0',
            ]),
        ),
    ]
)
async def test_device_register(
    expected_exception: AbstractContextManager[Exception],
    mock_device: DeviceMock
) -> None:
    """
    Tests for adding a new device to the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    with expected_exception:
        # Attempt to register a new device
        added_device = await g90.register_device(
            'Socket: S07', 'Test socket', timeout=1
        )

        # Verify sequence of commands sent to the panel
        assert await mock_device.recv_data == [
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
            b'ISTART[134,134,[134,'
            b'["Test socket",0,0,128,3,0,1,1190,0,17,1,0,2,"060A0600"]'
            b']]IEND\0',
            b'ISTART[162,162,[162,[1]]]IEND\0',
            b'ISTART[138,138,[138,[1,10]]]IEND\0',
        ]

        devices = await g90.devices
        # Verify the device list contains the new device
        assert len(devices) == 2
        new_device = devices[1]
        assert isinstance(new_device, G90Device)
        assert added_device is new_device
        assert new_device.name == "Test socket"
        assert new_device.index == 0
        assert new_device.type == 128


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTARTIEND\0',
    b'ISTART[162,[0]]IEND\0',
    b'ISTART[138,'
    b'[[1,1,2],'
    b'["Existing socket",11,0,128,3,0,32,0,0,16,1,0,""],'
    b'["Test socket",0,0,128,0,0,1,1480,0,17,4,0,""]'
    b']]IEND\0',
])
async def test_multinode_device_register(mock_device: DeviceMock) -> None:
    """
    Tests for adding a new multi-node device to the panel.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)
    # Attempt to register a new multi-node device
    await g90.register_device('Socket: JDQ', 'Test socket')

    # Verify sequence of commands sent to the panel
    assert await mock_device.recv_data == [
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
        b'ISTART[134,134,[134,'
        b'["Test socket",0,0,128,0,0,1,1480,0,17,4,0,2,"0707070B0B0D0D0E0E00"]'
        b']]IEND\0',
        b'ISTART[162,162,[162,[1]]]IEND\0',
        b'ISTART[138,138,[138,[1,10]]]IEND\0',
    ]

    devices = await g90.devices
    # Verify the device list contains the new multi-node device
    assert len(devices) == 5
    new_device = devices[1]
    assert isinstance(new_device, G90Device)
    assert devices[1].name == "Test socket#1"


@pytest.mark.g90device(sent_data=[
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
    b'ISTART[138,'
    b'[[1,1,1],["Switch",10,0,10,1,0,32,0,0,16,1,0,""]]]IEND\0',
])
async def test_device_update(mock_device: DeviceMock) -> None:
    """
    Tests updating the device and callback assocaited.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Set up the sensor list change callback to be called when the sensor
    # list is updated
    list_change_cb = MagicMock()
    g90.device_list_change_callback = list_change_cb
    future = asyncio.get_running_loop().create_future()
    list_change_cb.side_effect = (
        lambda *args: future.set_result(True)
    )

    # Trigger device list update
    devices = await g90.get_devices()

    # Verify the callback is called with correct parameters, indicating the
    # device has been added to the list
    await asyncio.wait([future], timeout=0.1)
    list_change_cb.assert_called_once_with(devices[0], True)

    # Subsequently retrieving the list of devices should result in same
    # callback invoked, but with different parameters
    future = asyncio.get_running_loop().create_future()
    await g90.get_devices()
    await asyncio.wait([future], timeout=0.1)
    list_change_cb.assert_called_with(devices[0], False)



"""
Tests for network configuration retrieval and modification.
"""
import pytest
from pyg90alarm.alarm import (
    G90Alarm,
)
from pyg90alarm.local.net_config import (
    G90NetConfig, G90APNAuth
)
from .device_mock import DeviceMock


@pytest.mark.g90device(sent_data=[
    b'ISTART[212,'
    b'[0,"123456789",1,1,"apn.a.net","user","pwd",3,"54321"]'
    b']IEND\0',
    b'ISTARTIEND\0'
])
async def test_net_config(mock_device: DeviceMock) -> None:
    """
    Tests for retrieving and modifying network configuration from
    the device.
    """
    g90 = G90Alarm(host=mock_device.host, port=mock_device.port)

    # Retrieve configuration
    cfg = await g90.net_config()
    assert isinstance(cfg, G90NetConfig)

    # Verify retrieved data
    assert cfg.ap_enabled is False
    assert cfg.ap_password == "123456789"
    assert cfg.wifi_enabled is True
    assert cfg.gprs_enabled is True
    assert cfg.apn_name == "apn.a.net"
    assert cfg.apn_user == "user"
    assert cfg.apn_password == "pwd"
    assert cfg.apn_auth == G90APNAuth.PAP_OR_CHAP
    assert cfg.gsm_operator == '54321'

    # Modify and save the configuration
    cfg.ap_enabled = True
    await cfg.save()

    # Verify data sent to the device
    assert await mock_device.recv_data == [
        b'ISTART[212,212,""]IEND\0',
        b'ISTART[213,213,[213,'
        b'[1,"123456789",1,1,"apn.a.net","user","pwd",3,"54321"]'
        b']]IEND\0'
    ]



"""
Tests for G90BaseList class
"""
from __future__ import annotations
from typing import AsyncGenerator
from unittest.mock import MagicMock

from pyg90alarm.entities.base_list import G90BaseList


class TestBaseList(G90BaseList[MagicMock]):
    """
    Mock subclass for testing G90BaseList.
    """

    # Prevent pytest from collecting this class as a test case
    __test__ = False

    async def _fetch(self) -> AsyncGenerator[MagicMock, None]:
        """Mock _fetch method."""
        if False:  # pragma: no cover
            yield MagicMock()


async def test_find_free_idx_empty_list() -> None:
    """
    Tests find_free_idx with empty entities list.
    """

    parent = MagicMock()
    base_list = TestBaseList(parent)
    base_list._entities = []

    result = await base_list.find_free_idx()

    assert result == 0


async def test_find_free_idx_one_entity_at_zero() -> None:
    """
    Tests find_free_idx with one entity at index 0.
    """

    base_list = TestBaseList(parent=MagicMock())
    base_list._entities = [MagicMock(index=0)]

    result = await base_list.find_free_idx()

    assert result == 1


async def test_find_free_idx_returns_lowest_available_index() -> None:
    """
    Tests find_free_idx with two entities at indexes 10 and 11.
    """

    base_list = TestBaseList(parent=MagicMock())
    base_list._entities = [MagicMock(index=10), MagicMock(index=11)]

    result = await base_list.find_free_idx()

    assert result == 0


async def test_find_free_idx_returns_lowest_available_index_over_gap() -> None:
    """
    Tests find_free_idx with two entities at indexes 10 and 11.
    """

    base_list = TestBaseList(parent=MagicMock())
    base_list._entities = [
        MagicMock(index=0), MagicMock(index=10), MagicMock(index=11)
    ]

    result = await base_list.find_free_idx()

    assert result == 1


async def test_find_entity_by_idx_and_name() -> None:
    """
    Tests find with matching index, subindex, and name.
    """

    base_list = TestBaseList(parent=MagicMock())
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = False
    base_list._entities = [entity]

    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result == entity


async def test_find_entity_name_mismatch() -> None:
    """
    Tests find with matching index but mismatched name.
    """

    base_list = TestBaseList(parent=MagicMock())
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = False
    base_list._entities = [entity]

    result = await base_list.find(
        idx=0, name="Wrong Name", exclude_unavailable=False
    )

    assert result is None


async def test_find_entity_not_found() -> None:
    """
    Tests find with non-existing index.
    """

    base_list = TestBaseList(parent=MagicMock())
    base_list._entities = []

    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result is None


async def test_find_entity_unavailable_excluded() -> None:
    """
    Tests find with unavailable entity and exclude_unavailable=True.
    """

    base_list = TestBaseList(parent=MagicMock())
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = True
    base_list._entities = [entity]

    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=True
    )

    assert result is None


async def test_find_entity_unavailable_not_excluded() -> None:
    """
    Tests find with unavailable entity and exclude_unavailable=False.
    """

    base_list = TestBaseList(parent=MagicMock())
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 0
    entity.name = "Test Entity"
    entity.is_unavailable = True
    base_list._entities = [entity]

    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False
    )

    assert result == entity


async def test_find_entity_with_subindex() -> None:
    """
    Tests find with specific subindex.
    """

    base_list = TestBaseList(parent=MagicMock())
    entity = MagicMock()
    entity.index = 0
    entity.subindex = 1
    entity.name = "Test Entity"
    entity.is_unavailable = False
    base_list._entities = [entity]

    result = await base_list.find(
        idx=0, name="Test Entity", exclude_unavailable=False, subindex=1
    )

    assert result == entity



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Protocol entity for G90 alarm panel phone numbers.
"""
from __future__ import annotations
from typing import Dict, Any
from dataclasses import dataclass
from ..const import G90Commands
from .dataclass_load_save import DataclassLoadSave


@dataclass
class G90AlarmPhones(DataclassLoadSave):
    """
    Interprets data fields of GETALMPHONE/SETALMPHONE commands.
    """
    # pylint: disable=too-many-instance-attributes
    LOAD_COMMAND = G90Commands.GETALMPHONE
    SAVE_COMMAND = G90Commands.SETALMPHONE

    # Password to operate the panel via SMS or incoming call.
    panel_password: str
    # Phone number of the alarm panel's SIM card.
    panel_phone_number: str
    # Alarm phone number to be called on alarm.
    # Should be in country code + number format.
    phone_number_1: str
    # Same, but for second alarm phone number.
    phone_number_2: str
    # Same, but for third alarm phone number.
    phone_number_3: str
    # Same, but for fourth alarm phone number.
    phone_number_4: str
    # Same, but for fifth alarm phone number.
    phone_number_5: str
    # Same, but for sixth alarm phone number.
    phone_number_6: str
    # Phone number to send SMS notifications on alarm.
    # Should be in country code + number format.
    sms_push_number_1: str
    # Same, but for second SMS notification phone number.
    sms_push_number_2: str

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the dataclass fields as a dictionary, masking sensitive data.
        """
        return {
            'panel_password': '********',
            'panel_phone_number': self.panel_phone_number,
            'phone_number_1': self.phone_number_1,
            'phone_number_2': self.phone_number_2,
            'phone_number_3': self.phone_number_3,
            'phone_number_4': self.phone_number_4,
            'phone_number_5': self.phone_number_5,
            'phone_number_6': self.phone_number_6,
            'sms_push_number_1': self.sms_push_number_1,
            'sms_push_number_2': self.sms_push_number_2,
        }



# Copyright (c) 2021 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
History protocol entity.
"""
import logging

from typing import Any, Optional, Dict
from dataclasses import dataclass
from datetime import datetime, timezone
from ..const import (
    G90AlertTypes,
    G90AlertSources,
    G90AlertStates,
    G90AlertStateChangeTypes,
    G90HistoryStates,
    G90RemoteButtonStates,
    G90RFIDKeypadStates,
)
from ..event_mapping import map_alert_state
from ..notifications.base import G90DeviceAlert

_LOGGER = logging.getLogger(__name__)


# The state of the incoming history entries are mixed of `G90AlertStates`,
# `G90AlertStateChangeTypes` and `G90RemoteButtonStates`, depending on entry
# type - hence separate dictionaries, since enums used for keys have
# conflicting values
states_mapping_alerts = {
    G90AlertStates.DOOR_CLOSE:
        G90HistoryStates.DOOR_CLOSE,
    G90AlertStates.DOOR_OPEN:
        G90HistoryStates.DOOR_OPEN,
    G90AlertStates.TAMPER:
        G90HistoryStates.TAMPER,
    G90AlertStates.LOW_BATTERY:
        G90HistoryStates.LOW_BATTERY,
    G90AlertStates.ALARM:
        G90HistoryStates.ALARM,
    G90AlertStates.MOTION_DETECTED:
        G90HistoryStates.MOTION_DETECTED,
}

states_mapping_state_changes = {
    G90AlertStateChangeTypes.AC_POWER_FAILURE:
        G90HistoryStates.AC_POWER_FAILURE,
    G90AlertStateChangeTypes.AC_POWER_RECOVER:
        G90HistoryStates.AC_POWER_RECOVER,
    G90AlertStateChangeTypes.DISARM:
        G90HistoryStates.DISARM,
    G90AlertStateChangeTypes.ARM_AWAY:
        G90HistoryStates.ARM_AWAY,
    G90AlertStateChangeTypes.ARM_HOME:
        G90HistoryStates.ARM_HOME,
    G90AlertStateChangeTypes.LOW_BATTERY:
        G90HistoryStates.LOW_BATTERY,
    G90AlertStateChangeTypes.WIFI_CONNECTED:
        G90HistoryStates.WIFI_CONNECTED,
    G90AlertStateChangeTypes.WIFI_DISCONNECTED:
        G90HistoryStates.WIFI_DISCONNECTED,
}

states_mapping_remote_buttons = {
    G90RemoteButtonStates.ARM_AWAY:
        G90HistoryStates.REMOTE_BUTTON_ARM_AWAY,
    G90RemoteButtonStates.ARM_HOME:
        G90HistoryStates.REMOTE_BUTTON_ARM_HOME,
    G90RemoteButtonStates.DISARM:
        G90HistoryStates.REMOTE_BUTTON_DISARM,
    G90RemoteButtonStates.SOS:
        G90HistoryStates.REMOTE_BUTTON_SOS,
}

states_mapping_rfid = {
    G90RFIDKeypadStates.ARM_AWAY:
        G90HistoryStates.RFID_KEY_ARM_AWAY,
    G90RFIDKeypadStates.ARM_HOME:
        G90HistoryStates.RFID_KEY_ARM_HOME,
    G90RFIDKeypadStates.DISARM:
        G90HistoryStates.RFID_KEY_DISARM,
    G90RFIDKeypadStates.LOW_BATTERY:
        G90HistoryStates.LOW_BATTERY,
    G90RFIDKeypadStates.CARD_0:
        G90HistoryStates.RFID_CARD_0,
    G90RFIDKeypadStates.CARD_1:
        G90HistoryStates.RFID_CARD_1,
    G90RFIDKeypadStates.CARD_2:
        G90HistoryStates.RFID_CARD_2,
    G90RFIDKeypadStates.CARD_3:
        G90HistoryStates.RFID_CARD_3,
    G90RFIDKeypadStates.CARD_4:
        G90HistoryStates.RFID_CARD_4,
}


@dataclass
class ProtocolData:
    """
    Class representing the data incoming from the device

    :meta private:
    """
    type: G90AlertTypes
    event_id: G90AlertStateChangeTypes
    source: G90AlertSources
    state: int
    sensor_name: str
    unix_time: int
    other: str


class G90History:
    """
    Represents a history entry from the alarm panel.
    """
    def __init__(self, *args: Any, **kwargs: Any):
        self._raw_data = args
        self._protocol_data = ProtocolData(*args, **kwargs)

    @property
    def datetime(self) -> datetime:
        """
        Date/time of the history entry.
        """
        return datetime.fromtimestamp(
            self._protocol_data.unix_time, tz=timezone.utc
        )

    @property
    def type(self) -> Optional[G90AlertTypes]:
        """
        Type of the history entry.
        """
        try:
            return G90AlertTypes(self._protocol_data.type)
        except (ValueError, KeyError):
            _LOGGER.warning(
                "Can't interpret '%s' as alert type (decoded protocol"
                " data '%s', raw data '%s')",
                self._protocol_data.type, self._protocol_data, self._raw_data
            )
            return None

    @property
    def state(self) -> Optional[G90HistoryStates]:
        """
        State for the history entry.
        """
        # pylint: disable=too-many-return-statements
        # No meaningful state for SOS alerts initiated by the panel itself
        # (host)
        if self.type == G90AlertTypes.HOST_SOS:
            return None

        try:
            # Remote button pressed or RFID keypad event occurred
            if self.type in [
                G90AlertTypes.SENSOR_ACTIVITY, G90AlertTypes.ALARM
            ]:
                # State of the remote indicate which button has been pressed
                if self.source == G90AlertSources.REMOTE:
                    return states_mapping_remote_buttons[
                        G90RemoteButtonStates(self._protocol_data.state)
                    ]

                # State of the RFID keypad indicate which action has occurred
                if self.source == G90AlertSources.RFID:
                    return states_mapping_rfid[
                        G90RFIDKeypadStates(self._protocol_data.state)
                    ]

            # Door open/close or alert types, mapped against `G90AlertStates`
            # using `state` incoming field
            if self.type in [
                G90AlertTypes.SENSOR_ACTIVITY, G90AlertTypes.ALARM
            ]:
                return G90HistoryStates(
                    # Map to history state via consolidated alert state
                    states_mapping_alerts[
                        # Map to consolidated alert state first
                        map_alert_state(
                            # Defaults to sensor source if none available
                            self.source or G90AlertSources.SENSOR,
                            self._protocol_data.state
                        )
                    ]
                )
        except (ValueError, KeyError):
            _LOGGER.warning(
                "Can't interpret '%s' as alert state (decoded protocol"
                " data '%s', raw data '%s')",
                self._protocol_data.state, self._protocol_data, self._raw_data
            )
            return None

        try:
            # Other types are mapped against `G90AlertStateChangeTypes`
            return G90HistoryStates(
                states_mapping_state_changes[
                    G90AlertStateChangeTypes(self._protocol_data.event_id)
                ]
            )
        except (ValueError, KeyError):
            _LOGGER.warning(
                "Can't interpret '%s' as state change (decoded protocol"
                " data '%s', raw data '%s')",
                self._protocol_data.event_id, self._protocol_data,
                self._raw_data
            )
            return None

    @property
    def source(self) -> Optional[G90AlertSources]:
        """
        Source of the history entry.
        """
        try:
            # Device state changes, open/close or alarm events are mapped
            # against `G90AlertSources` using `source` incoming field
            if self.type in [
                G90AlertTypes.STATE_CHANGE, G90AlertTypes.SENSOR_ACTIVITY,
                G90AlertTypes.ALARM
            ]:
                return G90AlertSources(self._protocol_data.source)
        except (ValueError, KeyError):
            _LOGGER.warning(
                "Can't interpret '%s' as alert source (decoded protocol"
                " data '%s', raw data '%s')",
                self._protocol_data.source, self._protocol_data, self._raw_data
            )
            return None

        # Other sources are assumed to be initiated by device itself
        return G90AlertSources.DEVICE

    @property
    def sensor_name(self) -> Optional[str]:
        """
        Name of the sensor related to the history entry, might be empty if none
        associated.
        """
        return self._protocol_data.sensor_name or None

    @property
    def sensor_idx(self) -> Optional[int]:
        """
        ID of the sensor related to the history entry, might be empty if none
        associated.
        """
        # Sensor ID will only be available if entry source is an infrared, RFID
        # keypad or other sensor
        if self.source in [
            G90AlertSources.SENSOR, G90AlertSources.RFID,
            G90AlertSources.INFRARED
        ]:
            return self._protocol_data.event_id

        return None

    def as_device_alert(self) -> G90DeviceAlert:
        """
        Returns the history entry represented as device alert structure,
        suitable for :meth:`G90DeviceNotifications._handle_alert`.
        """

        return G90DeviceAlert(
            type=self._protocol_data.type,
            event_id=self._protocol_data.event_id,
            source=self._protocol_data.source,
            state=self._protocol_data.state,
            zone_name=self._protocol_data.sensor_name,
            device_id='',
            unix_time=self._protocol_data.unix_time,
            resv4=0,
            other=self._protocol_data.other
        )

    def _asdict(self) -> Dict[str, Any]:
        """
        Returns the history entry as dictionary.
        """
        return {
            'type': self.type,
            'source': self.source,
            'state': self.state,
            'sensor_name': self.sensor_name,
            'sensor_idx': self.sensor_idx,
            'datetime': self.datetime,
        }

    def __repr__(self) -> str:
        """
        Textual representation of the history entry.
        """
        return super().__repr__() + f'({repr(self._asdict())})'



"""
Tests for G90BaseCommand class
"""
from unittest.mock import patch, DEFAULT
import re
import pytest

from pyg90alarm.local.base_cmd import (
    G90BaseCommand,
)
from pyg90alarm.exceptions import (
    G90Error, G90TimeoutError, G90CommandError, G90CommandFailure
)
from pyg90alarm.const import G90Commands

from .device_mock import DeviceMock


async def test_network_unreachable() -> None:
    """
    Verifies that network unreachable error is handled properly.
    """
    with patch.multiple(
        'socket', socket=DEFAULT, getaddrinfo=DEFAULT
    ) as mocks:
        g90 = G90BaseCommand(
            host='mocked', port=12345, code=G90Commands.GETHOSTINFO)

        # Simulate sending to device results in OS error
        mocks['socket'].return_value.send.side_effect = OSError(
            'Host unreachable'
        )
        mocks['getaddrinfo'].return_value = (5 * ('',),)

        with pytest.raises(OSError, match="Host unreachable"):
            await g90.process()


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_wrong_host(
    mock_device: DeviceMock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    Verifies that response from wrong host is handled properly.
    """
    orig = G90BaseCommand.datagram_received
    # Alter receving method of the device protocol as if it gets datagaram from
    # `another_host`
    monkeypatch.setattr(
        G90BaseCommand, 'datagram_received',
        lambda self, data, addr: orig(self, data, ('another_host', addr[1]))
    )
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=(
            'Received response from wrong host another_host,'
            f' expected from {mock_device.host}'
        )
    ):
        await g90.process()


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_wrong_port(
    mock_device: DeviceMock, monkeypatch: pytest.MonkeyPatch
) -> None:
    """
    Verifies that response from wrong port is handled properly.
    """
    orig = G90BaseCommand.datagram_received
    # Alter receving method of the device protocol as if it gets datagaram from
    # proper host but different port `54321`
    monkeypatch.setattr(
        G90BaseCommand, 'datagram_received',
        lambda self, data, addr: orig(self, data, (addr[0], 54321))
    )
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=(
            'Received response from wrong port 54321,'
            f' expected from {mock_device.port}'
        )
    ):
        await g90.process()


# No data the simulated device sends back will result in receive timeout for
# the client
@pytest.mark.g90device(sent_data=[])
async def test_timeout(mock_device: DeviceMock) -> None:
    """
    Verifies that timeout is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO,
        timeout=0.1, retries=2
    )

    with pytest.raises(G90TimeoutError):
        await g90.process()
    assert await mock_device.recv_data == [
        b'ISTART[206,206,""]IEND\0',
        b'ISTART[206,206,""]IEND\0',
    ]


@pytest.mark.g90device(sent_data=[
    b'\xdeadbeef\0',
])
async def test_invalid_utf8_encoding(mock_device: DeviceMock) -> None:
    """
    Verifies that invalid UTF-8 encoding of response is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=re.escape("Unable to decode response from UTF-8")
    ):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTART[IEND\0',
])
async def test_wrong_format(mock_device: DeviceMock) -> None:
    """
    Verifies that wrong format of response is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=re.escape("Unable to parse response as JSON: '['")
    ):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTARTIEND\0',
])
async def test_empty_response(mock_device: DeviceMock) -> None:
    """
    Verifies that empty response is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTART[]IEND\0',
])
async def test_no_code_response(mock_device: DeviceMock) -> None:
    """
    Verifies that response without code is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=re.escape("Missing code in response: '[]'")
    ):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTART[106,[""]]IEND\0',
])
async def test_wrong_code_response(mock_device: DeviceMock) -> None:
    """
    Verifies that response with wrong code is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match='Wrong response - received code 106, expected code 206'
    ):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTART[206]IEND\0',
])
async def test_no_data_response(mock_device: DeviceMock) -> None:
    """
    Verifies that response without data is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90Error,
        match=re.escape("Missing data in response: '[206]'")
    ):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'dummy',
])
async def test_no_start_marker(mock_device: DeviceMock) -> None:
    """
    Verifies that response without start marker is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(G90Error, match='Missing start marker in data'):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


@pytest.mark.g90device(sent_data=[
    b'ISTART[206,[]]IEND',
])
async def test_no_end_marker(mock_device: DeviceMock) -> None:
    """
    Verifies that response without end marker is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(G90Error, match='Missing end marker in data'):
        await g90.process()
    assert await mock_device.recv_data == [b'ISTART[206,206,""]IEND\0']


async def test_command_code_none_error(mock_device: DeviceMock) -> None:
    """
    Verifies that using `NONE` command code is disallowed by `G90BaseCommand`
    class.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.NONE
    )

    with pytest.raises(G90Error, match="'NONE' command code is disallowed"):
        await g90.process()


@pytest.mark.g90device(sent_data=[
    b'ISTARTfailIEND\0',
])
async def test_command_failure(mock_device: DeviceMock) -> None:
    """
    Verifies that command failure is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90CommandFailure,
        match=re.escape('Command GETHOSTINFO (code=206) failed')
    ):
        await g90.process()


@pytest.mark.g90device(sent_data=[
    b'ISTARTerrordummyIEND\0',
])
async def test_command_error(mock_device: DeviceMock) -> None:
    """
    Verifies that command error is handled properly.
    """
    g90 = G90BaseCommand(
        host=mock_device.host, port=mock_device.port,
        code=G90Commands.GETHOSTINFO
    )

    with pytest.raises(
        G90CommandError,
        match=re.escape(
            "Command GETHOSTINFO (code=206) failed with error: 'dummy'"
        )
    ):
        await g90.process()



# Copyright (c) 2022 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""
Sensor definitions for G90 alarm panel.
"""
from __future__ import annotations
import logging
from .base import (
    G90PeripheralDefinition,
    G90PeripheralDefinitionsBase,
    G90PeripheralTypes,
    G90PeripheralProtocols,
    G90PeripheralRwModes,
    G90PeripheralMatchModes,
    unique_definitions
)

_LOGGER = logging.getLogger(__name__)


@unique_definitions
class G90SensorDefinitions(G90PeripheralDefinitionsBase):
    """
    Sensor definitions for G90 devices, required when modifying them since
    writing a sensor to the device requires values not present on read.
    """
    SENSOR_DEFINITIONS = [
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Door Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=2,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Glass Break Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=3,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Gas Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=4,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Smoke Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=5,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="SOS Button: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=6,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Vibration Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=7,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Water Leak Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=9,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Beam Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=8,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="PIR Motion Sensor: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=11,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="RFID: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.CORD_SENSOR,
            subtype=12,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Door Bell: Wired",
            protocol=G90PeripheralProtocols.CORD,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.IR_2_4G,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Infrared Transcoder",
            protocol=G90PeripheralProtocols.RF_PRIVATE,
            timeout=0,
            baudrate=800,
            node_count=1
        ),

        G90PeripheralDefinition(
            type=G90PeripheralTypes.TOUCH_SWITCH_2_4G,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Switch: 2.4G Type 1",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.TOUCH_SWITCH_2_4G,
            subtype=2,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Switch: 2.4G Type 2",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=2
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.TOUCH_SWITCH_2_4G,
            subtype=3,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Switch: 2.4G Type 3",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=3
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.TOUCH_SWITCH_2_4G,
            subtype=4,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.WRITE,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Switch: 2.4G Type 4",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=960,
            node_count=4
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.DOOR,
            subtype=0,
            rx=5,
            tx=0,
            private_data='F100f5017d02f803f90400',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Door Sensor: WDS07",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.DOOR,
            subtype=1,
            rx=2,
            tx=0,
            private_data='0001020400',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Door Sensor: Non-smart",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.DOOR,
            subtype=3,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Door Sensor: WRDS01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.DOOR,
            subtype=2,
            rx=5,
            tx=0,
            private_data='F100f5017d02f803f904',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Door Sensor: Smart",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.GLASS,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Glass Break Sensor: BLPS",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.GAS,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Gas Sensor: WGD01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SMOKE,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Smoke Sensor: WSD02",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SMOKE,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Smoke Sensor: WSD04",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOS,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="SOS Button: WEB01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SOS,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="SOS Button: WEB03",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.VIB,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Vibration Sensor: WSS01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.WATER,
            subtype=1,
            rx=0,
            tx=0,
            private_data='',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Water Leak Sensor: LSTC02",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.WATER,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Water Leak Sensor: LSTC01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=3,
            rx=3,
            tx=0,
            private_data='06000301020200',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="PIR Motion Sensor: WMS08",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=12,
            rx=3,
            tx=0,
            private_data='060003010202000314',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="PIR Motion Sensor: WMS08B",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: ODPIR",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=5,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: N650",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=6,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: WPD02",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=8,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: WCMS02",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=9,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: CWMS01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=11,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: WMS04",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=2,
            rx=3,
            tx=0,
            private_data='04000801020200',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="PIR Motion Sensor: WMS07",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=4,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: ODPIR03",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=7,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: WPD01",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.INFRARED,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="PIR Motion Sensor: Ceiling",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.IN_BEAM,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Beam Sensor: ABT",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.IN_BEAM,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Beam Sensor: ABE",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.IN_BEAM,
            subtype=2,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Beam Sensor: ABH",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.REMOTE,
            subtype=1,
            rx=4,
            tx=0,
            private_data='0d000b010e02070300',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Remote: RMC08",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.REMOTE,
            subtype=0,
            rx=4,
            tx=0,
            private_data='cf00fc01f3023f0300',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Remote: RMC07",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.REMOTE,
            subtype=4,
            rx=4,
            tx=0,
            private_data='cf00fc01f3023f0300',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Remote: RMC02",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.REMOTE,
            subtype=2,
            rx=3,
            tx=0,
            private_data='07000602050300',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Remote: 3 Button",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.RFID,
            subtype=1,
            rx=11,
            tx=0,
            private_data='0e0007010d020b03010402050c060a0709080809060a00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="RFID: K07",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.DOORBELL,
            subtype=1,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Door Bell: WDB",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.BUTTONID,
            subtype=1,
            rx=4,
            tx=0,
            private_data='07000e010d020b0300',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Fingerprint: Sensor",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.WATCH,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="SOS Button: Watch",
            protocol=G90PeripheralProtocols.RF_2262,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.FINGER_LOCK,
            subtype=1,
            rx=0,
            tx=2,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY16BITS,
            name="Fingerprint: Lock",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.SUBHOST,
            subtype=0,
            rx=0,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ALL,
            name="Sub Host: SS08S",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1480,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.GAS_VALVE,
            subtype=0,
            rx=2,
            tx=1,
            private_data='09000801ff0e0e',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Gas Valve Sensor: WGD02",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=1600,
            node_count=1
        ),
        G90PeripheralDefinition(
            type=G90PeripheralTypes.REMOTE_2_4G,
            subtype=4,
            rx=4,
            tx=0,
            private_data='00',
            rw_mode=G90PeripheralRwModes.READ,
            match_mode=G90PeripheralMatchModes.ONLY20BITS,
            name="Remote: RMC2.4G",
            protocol=G90PeripheralProtocols.RF_1527,
            timeout=0,
            baudrate=0,
            node_count=1
        ),
    ]

    @classmethod
    def definitions(cls) -> list[G90PeripheralDefinition]:
        """
        Gets all sensor definitions.

        :return: List of sensor definitions.
        """
        _LOGGER.debug(
            "Number of sensor definitions: %d", len(cls.SENSOR_DEFINITIONS)
        )
        return cls.SENSOR_DEFINITIONS



# Copyright (c) 2026 Ilia Sotnikov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Interprets network configuration data fields of GETAPINFO/SETAPINFO commands.
"""
from __future__ import annotations
from enum import IntEnum
from dataclasses import dataclass
from typing import Any, Dict
from ..const import G90Commands
from .dataclass_load_save import DataclassLoadSave


class G90APNAuth(IntEnum):
    """
    Supported APN authentication methods.
    """
    NONE = 0
    PAP = 1
    CHAP = 2
    PAP_OR_CHAP = 3


@dataclass
class G90NetConfig(DataclassLoadSave):
    """
    Interprets data fields of GETAPINFO/SETAPINFO commands.
    """
    # pylint: disable=too-many-instance-attributes
    LOAD_COMMAND = G90Commands.GETAPINFO
    SAVE_COMMAND = G90Commands.SETAPINFO

    # Whether the access point is enabled, so that the device can be accessed
    # via WiFi
    _ap_enabled: int
    # Access point password
    ap_password: str
    # Whether WiFi is enabled, so that the device can connect to WiFi network
    _wifi_enabled: int
    # Whether GPRS is enabled, so that the device can connect via cellular
    # network
    _gprs_enabled: int
    # Access Point Name (APN) for GPRS connection, as provided by the cellular
    # operator
    apn_name: str
    # User name for APN authentication, as provided by the cellular operator
    apn_user: str
    # Password for APN authentication, as provided by the cellular operator
    apn_password: str
    # APN authentication method, as provided by the cellular operator
    _apn_auth: int
    # GSM operator code
    gsm_operator: str

    @property
    def ap_enabled(self) -> bool:
        """
        Returns whether the access point is enabled.
        """
        return bool(self._ap_enabled)

    @ap_enabled.setter
    def ap_enabled(self, value: bool) -> None:
        self._ap_enabled = int(value)

    @property
    def wifi_enabled(self) -> bool:
        """
        Returns whether WiFi is enabled.
        """
        return bool(self._wifi_enabled)

    @wifi_enabled.setter
    def wifi_enabled(self, value: bool) -> None:
        self._wifi_enabled = int(value)

    @property
    def gprs_enabled(self) -> bool:
        """
        Returns whether GPRS is enabled.
        """
        return bool(self._gprs_enabled)

    @gprs_enabled.setter
    def gprs_enabled(self, value: bool) -> None:
        self._gprs_enabled = int(value)

    @property
    def apn_auth(self) -> G90APNAuth:
        """
        Returns the APN authentication method as an enum.
        """
        return G90APNAuth(self._apn_auth)

    @apn_auth.setter
    def apn_auth(self, value: G90APNAuth) -> None:
        self._apn_auth = value.value

    def _asdict(self) -> Dict[str, Any]:
        return {
            'ap_enabled': self.ap_enabled,
            'ap_password': '********',
            'wifi_enabled': self.wifi_enabled,
            'gprs_enabled': self.gprs_enabled,
            'apn_name': self.apn_name,
            'apn_user': self.apn_user,
            'apn_password': '********',
            'apn_auth': self.apn_auth,
            'gsm_operator': self.gsm_operator
        }
